"""
Convenience functions for registration related image methods

Author: Fabian A. Preiss.
"""
from typing import Dict, Hashable, Sequence, Tuple, cast, Optional

import numpy as np
from numpy.linalg import norm
from scipy.fftpack import fft2, fftshift
from skimage.filters import difference_of_gaussians, window
from skimage.registration import phase_cross_correlation
from skimage.transform import AffineTransform, warp, warp_polar, radon


class ImageMethods:
    """Collection of static methods for image analysis and manipulation."""

    @staticmethod
    def norm_rel_l2(image: np.ndarray, ref_image: np.ndarray) -> float:
        r"""
        Compute a relative similarity measurement between two images.

        Interpretes the images as a vector and calculates the L2 norm of the differences relative to the
        reference image `ref_image`.

        Notes
        -----
        L2 norm  Implemented analog to `NormRel_L2` [#f1]_.

        .. math::
            \frac{\left\Vert \mathtt{ref\_image}-\mathtt{image}\right\Vert _{F}}{\left\Vert \mathtt{ref\_image}\right\Vert _{F}}

        Where :math:`\left\Vert \ldots\right\Vert _{F}` denotes the Frobenius norm of a matrix.

        References
        ----------
        .. [#f1] `NVIDIA Performance Primitives (NPP) - Image Norms
                 <https://docs.nvidia.com/cuda/archive/9.2/npp/group__image__norm.html>`_
        """
        return cast(float, norm(ref_image - image) / norm(ref_image))

    @staticmethod
    def abs_diff(image: np.ndarray, ref_image: np.ndarray) -> np.ndarray:
        r"""
        Absolute value of the difference between two images.

        Notes
        -----
        .. math::
            \mathtt{abs}\left(\mathtt{ref\_image}-\mathtt{image}\right)
        """
        return cast(np.ndarray, abs(ref_image - image))

    @staticmethod
    def sqr_diff(image: np.ndarray, ref_image: np.ndarray) -> np.ndarray:
        r"""
        Squared difference between two images.

        Notes
        -----
        .. math::
            \left(\mathtt{ref\_image}-\mathtt{image}\right)^{2}
        """
        return cast(np.ndarray, (ref_image - image) ** 2)

    @staticmethod
    def exp_filter(
        image: np.ndarray, signal_noise_ratio: Optional[float] = None
    ) -> np.ndarray:
        """
        Remap the values of the image such that bright pixels are given an exponentially higher weight.

        Maps the min value of the input image to 1/signal_noise_ratio and the max value to 1.
        """
        if signal_noise_ratio is not None:
            norm_image = np.interp(
                image, (image.min(), image.max()), (-np.log(signal_noise_ratio), 0)
            )
            return cast(np.ndarray, np.exp(norm_image))
        return image

    @staticmethod
    def sinogram(image, theta=None, exp_filter_val=None, circle=False) -> np.ndarray:
        """Computes the radon transformation and optionally applies the `exp_filter` afterwards.

        Examples
        --------
        .. plot:: tutorial/pyplots/sinogram.py
           :include-source:
        """
        sinogram = radon(image, theta=theta, circle=circle)
        if exp_filter_val is not None:
            sinogram = ImageMethods.exp_filter(sinogram, exp_filter_val)
        return cast(np.ndarray, sinogram)

    @staticmethod
    def sinogram_project(
        image, theta=None, exp_filter_val=None, circle=False
    ) -> np.ndarray:
        """Projects the sinogram onto the axis of the theta angle.

        Examples
        --------
        .. plot:: tutorial/pyplots/sinogram_project.py
           :include-source:
        """
        sinogram = ImageMethods.sinogram(
            image, theta=theta, exp_filter_val=exp_filter_val, circle=circle
        )
        projection = np.sum(sinogram, axis=0)
        return cast(np.ndarray, projection)

    @staticmethod
    def max_sinogram_angle(
        image, theta=None, exp_filter_val=None, circle=False, precision=0.1
    ) -> float:
        """Iterative solver, works well in special cases. Implementation very crude."""
        if theta is None:
            theta = np.arange(0, 180, 1)
        projection = ImageMethods.sinogram_project(
            image, theta=theta, exp_filter_val=exp_filter_val, circle=circle
        )
        max_index = np.where(np.max(projection) == projection)
        angles = theta[max_index]
        if len(max_index[0]) > 1:
            print(
                f"Warning: angle solution not unique returning first match, possible angles: {angles}"
            )
        delta_theta = theta[1] - theta[0]
        if delta_theta > precision:
            theta = np.linspace(
                angles[0] - delta_theta * 2, angles[0] + delta_theta * 2, 10
            )
            return ImageMethods.max_sinogram_angle(
                image,
                theta=theta,
                exp_filter_val=exp_filter_val,
                precision=precision,
                circle=circle,
            )
        return cast(float, angles[0] % 180)

    @staticmethod
    def compute_rts(
        image: np.ndarray,
        angle: float = 0,
        scale: float = 1,
        translation: Sequence[float] = (0.0, 0.0),
        inverse: bool = False,
        preserve_range: bool = True,
        order: int = 5,
    ) -> np.ndarray:
        """
        Rotate, translate and scale image.

        Parameters
        ----------
        image : numpy.ndarray
            The input image to transform
        angle : float, optional
            The rotation angle in degrees for the transform
        scale : float, optional
            The scaling factor used in the transform
        translation : (float, float), optional
            x,y-translations used in the transform
        inverse : bool
            Apply the backwards transformation for given parameters

        Returns
        -------
        numpy.ndarray
            modified image with same shape as initial image
        """
        # AffineTransform rotates around (x,y)=(0,0) instead of
        # image center => calculate coordinate transformation
        _translation = np.array(translation, dtype="float64")
        tform = AffineTransform(scale=scale, rotation=np.deg2rad(angle))
        # change in ordering required, as numpy.ndarray uses matrix-like row/column indexing
        shape_half = (np.array(image.shape) / 2)[::-1]
        _translation = shape_half + tform.params[:2, :2] @ (_translation - shape_half)
        tform.translation[:] = _translation
        if not inverse:
            rts_image = warp(
                image, tform, order=order, preserve_range=preserve_range
            )  # ,mode='wrap')
        else:
            rts_image = warp(
                image, tform.inverse, order=order, preserve_range=preserve_range
            )  # ,mode='wrap')
        return cast(np.ndarray, rts_image)

    @staticmethod
    def compute_dgfw(
        image: np.ndarray,
        gaussdiff: Sequence[float] = (5, 20),
        windowweight: float = 1,
        windowtype: str = "hann",
    ) -> np.ndarray:
        """Image Difference of Gaussian Filter + Window.

        Parameters
        ----------
        image : numpy.ndarray
            The input image to filter
        gaussdiff : (float, float), optional
            The low and high standard deviations for the gaussian difference band pass filter
        windowweight : float, optional
            weighting factor scaling beween windowed image and image
        windowtype : str, optional
            see `skimage.filters.window` for possible choices

        Returns
        -------
        numpy.ndarray
            modified image with bandpass filter and window applied

        Notes
        -----
        Applying this bandpass and window filter prevents artifacts from image boundaries and
        noise from contributing significantly to the fourier transform. The gaussian
        difference filter can be tuned such that the features relevant for the identification
        of the rotation angle are at the center of the band pass filter.
        """
        image_dgfw = (
            difference_of_gaussians(image, *gaussdiff)
            if gaussdiff[0] < gaussdiff[1]
            else image * 1.0
        )
        image_dgfw *= windowweight * window(windowtype, image.shape) + (
            1 - windowweight
        )
        return cast(np.ndarray, image_dgfw)

    @staticmethod
    def compute_afts(image: np.ndarray) -> np.ndarray:
        """Compute FFT magnitude, shifted with low fequencies in center.

        Parameters
        ----------
        image : numpy.ndarray
            The input image for the fourier transform

        Returns
        -------
        numpy.ndarray
            FFT magnitude, center shifted
        """
        image_fs = np.abs(fftshift(fft2(image)))
        return cast(np.ndarray, image_fs)

    @staticmethod
    def compute_log_polar_tf(
        image: np.ndarray, wrexp: float = 3, order: int = 5
    ) -> np.ndarray:
        """Compute log-scaled polar coordinate transform of center shifted FFT.

        Parameters
        ----------
        image : numpy.ndarray
            The input image to transform (expects center shifted FFT magnitude)
        wrexp : float, optional
            Cutoff exponent factor for higher frequencies, larger wrexp => faster computation
            min value: 1

        Returns
        -------
        numpy.ndarray
            log-scaled polar transformed of the input image
        """
        warp_radius = ImageMethods.compute_warp_radius(min(image.shape), wrexp)
        # NOTE: we can use 5th order approximation here, as a the computation time on our image
        # only increases lineary with the interpolation order, whereas the mean squared error
        # decreases pretty much exponentially
        warped = warp_polar(
            image,
            radius=warp_radius,
            output_shape=image.shape,
            scaling="log",
            order=order,
        )
        # FFT magnitude is symmetric => div by 2
        return cast(np.ndarray, warped[: image.shape[0] // 2, :])

    @staticmethod
    def compute_warp_radius(image_diameter: int, wrexp: float = 1.0) -> int:
        """Compute the warp radius from image and warp radius exponent *wrexp*.

        Parameters
        ----------
        image_diameter : int
            The length of the smallest image dimension
        wrexp : float, optional
            Cutoff exponent factor for higher frequencies, larger wrexp => faster computation
            min value: 1

        Returns
        -------
        int
            The cutoff radius for the log-ploar transform of the image
        """
        return int(image_diameter // (2 ** max(wrexp, 1.0)))

    # NOTE: ugly return params
    @staticmethod
    def recover_rs(
        image_warped_fs: np.ndarray,
        rts_warped_fs: np.ndarray,
        image_shape: Sequence[int],
        upsampl: int = 10,
        wrexp: float = 3,
    ) -> Tuple[np.ndarray, np.ndarray, Dict[str, Hashable]]:
        """Recover the rotation and scaling transformation from given input.

        Parameters
        ----------
        image_warped_fs : np.ndarray
            log-polar warped fourier transformed of original input image
        rts_warped_fs : np.ndarray
            log-polar warped fourier transformed of modified input image
        image_shape : Sequence[int]
            image dimensions of original input image
        upsampl : int, optional
            Upsampling factor. 1 => no upsampling, 20 => precision to 1/20 of a pixel
        wrexp : float, optional
            Cutoff exponent factor for higher frequencies, larger wrexp => faster computation
            min value: 1

        Returns
        -------
        numpy.ndarray
            Vector of recovered rotation angle and error in degrees
        numpy.ndarray
            Vector of recovered scaling factor and error
        dict
            Dict containing the phase_cross_correlation parameters

        Notes
        -----
        The errors are a lower estimate under ideal assumptions and can be much larger depending on the data.
        """
        warp_radius = ImageMethods.compute_warp_radius(min(image_shape), wrexp)
        shifts, error, phasediff = phase_cross_correlation(
            image_warped_fs, rts_warped_fs, upsample_factor=upsampl
        )
        phaseparams = {"shifts": shifts, "error": error, "phasediff": phasediff}
        shift_err_r = np.array([shifts[0], error])
        shift_err_c = np.array([shifts[1], error])
        recovered_rotation = cast(np.ndarray, (360 / image_shape[0]) * shift_err_r)
        klog = image_shape[1] / np.log(warp_radius)
        recovered_scale = cast(np.ndarray, np.exp(shift_err_c / klog))
        return recovered_rotation, recovered_scale, phaseparams



"""
Enum of the validator.

Author: Fabian A. Preiss
"""
from enum import Enum


class AngleSelectParams(Enum):
    IMG = "IMG"
    REF_IMG = "REF_IMG"
    ANGLE_A = "ANGLE_A"
    ANGLE_B = "ANGLE_B"
    UPSAMPLING = "UPSAMPLING"
    ROT_A_IMG = "ROT_A_IMG"
    ROT_B_IMG = "ROT_B_IMG"
    TRANSLATION_A = "TRANSLATION_A"
    TRANSLATION_B = "TRANSLATION_B"
    ROT_TR_A_IMG = "ROT_TR_A_IMG"
    ROT_TR_B_IMG = "ROT_TR_B_IMG"
    SELECTOR = "SELECTOR"
    RECOVERED_ROTATION = "RECOVERED_ROTATION"
    RECOVERED_ROT_IMG = "RECOVERED_ROT_SCALE_IMG"
    RECOVERED_TRANSLATION = "RECOVERED_TRANSLATION"
    RECOVERED_ROT_TR_IMG = "RECOVERED_ROT_TR_IMG"



"""
Image registration based on the radon transform.

Author: Fabian A. Preiss
"""

import numpy as np
from imgreg.models.radon.params import (
    RefImageParam,
    ModImageParam,
    AngularPrecisionParam,
    ExponentialFilterSignalNoiseParam,
    ThetaParam,
    UpsamplingParam,
    RefRotationParam,
    ModRotationParam,
    RotationCandidateParam,
    AngleSelectParam,
    RecoveredRotationParam,
    RecoveredRotParam,
    RecoveredTranslationParam,
    RecoveredRotTrParam,
)
from imgreg.models.radon.enums import RadonParams
from imgreg.util.solver import Solver
from typing import Optional


class RadonSolver(Solver):
    r"""
    Implements an image registration model based on the radon transform.

    The model tries to reconstruct the difference of rotation and translation between two images.
    The model is limited such, that due to the symmetry of the method it only considers
    angles between 0 and 180 degrees. It will fail, if the rotation angle is larger.

    Parameters
    ----------
    ref_img : numpy.ndarray
        The original input image (one color channel only).
    mod_img : numpy.ndarray
        The modified input image (one color channel only).

    Notes
    -----
    The model implements the following dependency graph to construct it's `Parameter`\ s.

        .. graphviz:: ../_assets/dependencies_radon.gv

    The `Parameter`\ s are documented in :ref:`imgreg.models.radon.params`.

    Examples
    --------
    We can visualize the internal `ImageParameter`\ s of the model as follows:

    .. plot:: tutorial/pyplots/radonsolver.py
       :include-source:

    If we simply want to create a model and access the recovered values we first setup a model:

    >>> import numpy as np
    >>> import imgreg.data as data
    >>> from imgreg.models.radon import RadonSolver
    >>> ref_img = np.array(data.ref_img())
    >>> mod_img = np.array(data.mod_img())
    >>> ras = RadonSolver(ref_img, mod_img)

    Now the parameters of the model can now be accessed as follows:

    >>> ras.RECOVERED_ROTATION.value
    array([-11.35802469,   0.28284271])

    >>> ras.RECOVERED_TRANSLATION.value
    array([-11.80822224,  25.77936941,   0.30466765])
    """

    def __init__(
        self, ref_img: Optional[np.ndarray] = None, mod_img: Optional[np.ndarray] = None
    ):
        super().__init__()
        p_ref_img = RefImageParam(ref_img)
        p_mod_img = ModImageParam(mod_img)
        p_angular_precision = AngularPrecisionParam(0.2)
        p_exponential_filter_signal_noise = ExponentialFilterSignalNoiseParam(100)
        p_theta = ThetaParam(np.arange(0, 180, 1))
        p_upsampling = UpsamplingParam(10)
        p_ref_rotation = RefRotationParam(
            {p_ref_img, p_theta, p_exponential_filter_signal_noise, p_angular_precision}
        )
        p_mod_rotation = ModRotationParam(
            {p_mod_img, p_theta, p_exponential_filter_signal_noise, p_angular_precision}
        )
        p_rotation_candidate = RotationCandidateParam({p_ref_rotation, p_mod_rotation})
        p_angle_select = AngleSelectParam(
            {p_mod_img, p_ref_img, p_rotation_candidate, p_upsampling}
        )

        p_recovered_rotaton = RecoveredRotationParam(
            {p_rotation_candidate, p_angle_select}
        )
        p_recovered_rot = RecoveredRotParam({p_angle_select})
        p_recovered_translation = RecoveredTranslationParam({p_angle_select})
        p_recovered_rot_tr = RecoveredRotTrParam({p_angle_select})
        self._register_params(
            {
                p_ref_img,
                p_mod_img,
                p_angular_precision,
                p_exponential_filter_signal_noise,
                p_theta,
                p_upsampling,
                p_ref_rotation,
                p_mod_rotation,
                p_rotation_candidate,
                p_angle_select,
                p_recovered_rotaton,
                p_recovered_rot,
                p_recovered_translation,
                p_recovered_rot_tr,
            }
        )
        self._generate_dep_graphs()
        for key in RadonParams:
            setattr(self, key.name, self[key])


#



"""
Enum of the logpolar parameters.

Author: Fabian A. Preiss
"""
from enum import Enum


class LogPolParams(Enum):
    GAUSS_DIFF = "GAUSS_DIFF"
    WINDOW_WEIGHT = "WINDOW_WEIGHT"
    WINDOW_TYPE = "WINDOW_TYPE"
    WINDOW_RADIUS_EXP = "WINDOW_RADIUS_EXP"
    BOUNDS = "BOUNDS"
    REF_IMG = "REF_IMG"
    MOD_IMG = "MOD_IMG"
    GAUSS_DIFF_REF_IMG = "GAUSS_DIFF_REF_IMG"
    GAUSS_DIFF_MOD_IMG = "GAUSS_DIFF_MOD_IMG"
    FOURIER_REF_IMG = "FOURIER_REF_IMG"
    FOURIER_MOD_IMG = "FOURIER_MOD_IMG"
    WARPED_FOURIER_REF_IMG = "WARPED_FOURIER_REF_IMG"
    WARPED_FOURIER_MOD_IMG = "WARPED_FOURIER_MOD_IMG"
    UPSAMPLING = "UPSAMPLING"
    RECOVERED_ROTATION_SCALE_PHASE = "RECOVERED_ROTATION_SCALE_PHASE"
    RECOVERED_ROTATION = "RECOVERED_ROTATION"
    RECOVERED_SCALE = "RECOVERED_SCALE"
    RECOVERED_TRANSLATION = "RECOVERED_TRANSLATION"
    RECOVERED_ROT_SCALE_IMG = "RECOVERED_ROT_SCALE_IMG"
    RECOVERED_ROT_SCALE_TR_IMG = "RECOVERED_ROT_SCALE_TR_IMG"



from imgreg.util.graph import DAGraph
from imgreg.util.io import fnmatch_filter, dirtree_reader, File_Set_Ops, DirectoryView
from imgreg.util.params import (
    ParameterError,
    interface_function_handle,
    Parameter,
    ImageParameter,
)
from imgreg.util.solver import SolverError, Solver
from imgreg.util.methods import ImageMethods
from imgreg.util.helpers import image_file_gen, solver_gen



"""
Parameter classes implementing lazy evaluation and dependency resolution.

Author: Fabian A. Preiss
"""

from __future__ import annotations

import types
from enum import Enum
from typing import Any, Callable, Dict, Optional, Set, Tuple, Union, Type, cast

import matplotlib.pyplot as plt
import types
from typeguard import check_type


class ParameterError(Exception):
    pass


class ParameterRegisterError(Exception):
    pass


def interface_function_handle(parameter: Dict[Enum, Parameter]) -> Any:
    r"""Construct value using parent `Parameter`\ s"""
    raise NotImplementedError


# TODO
# parents/children internal access: Dict[Enum, Parameter]
# parents/children external access: Set[Parameter]
class Parameter:
    r"""
    Base `Parameter` class, aware of related `Parameter`\ s and lazy evaluating.
    """

    def __init__(
        self,
        enum_id: Enum,
        value_definition: Union[Type[Any], Callable[[Dict[Enum, Parameter]], Any]],
        parent_parameters: Optional[Set[Parameter]] = None,
    ):
        self.__enum_id = enum_id
        self.__parents: Dict[Enum, Parameter] = (
            {}
            if parent_parameters is None
            else {parent.enum_id: parent for parent in parent_parameters}
        )
        # children have to be set from outside user
        self.__children: Dict[Enum, Parameter] = {}
        self.__descendants: Dict[Enum, Parameter] = {}
        self.__function_handle: Callable[[Dict[Enum, Parameter]], Any]
        self.__type_info: Union[Type[Any], Callable[[Dict[Enum, Parameter]], Any]]

        if isinstance(value_definition, types.FunctionType):
            self.__type_info = value_definition.__annotations__.get("return", Any)
            self.__function_handle = value_definition
        else:
            try:
                self.__type_info = value_definition
            except TypeError as err:
                raise ParameterError(
                    "value_definition is neither a type or a function."
                ) from err
            self.__function_handle = interface_function_handle
        if self.__type_info is Any:
            raise ParameterError(
                "value_definition requires type informations other than Any."
            )
        self.__value: Any = None
        self.__const: bool = False

    def __gt__(self, other: Parameter) -> bool:
        if type(self).__name__ > type(other).__name__:
            return True
        if type(self).__name__ < type(other).__name__:
            return False
        return self.enum_id.name > other.enum_id.name

    def __lt__(self, other: Parameter) -> bool:
        if type(self).__name__ > type(other).__name__:
            return False
        if type(self).__name__ < type(other).__name__:
            return True
        return self.enum_id.name < other.enum_id.name

    @property
    def enum_id(self) -> Enum:
        """Returns the Parameter ID as an enumeration."""
        return self.__enum_id

    @property
    def type_info(self):
        """The type of this `Parameter`."""
        return self.__type_info

    @property
    def const(self) -> bool:
        """Flag if the value can be overwritten once initialized."""
        return self.__const

    @const.setter
    def const(self, const: bool) -> None:
        if self.__const:
            raise ParameterError("Parameter is already flagged as constant.")
        self.__const = const

    @property
    def value(self):
        """The value of this `Parameter`."""
        if self.__value is None:
            self.__clear_descendants()
            try:
                self.__value = self.__function_handle(self.__parents)
            except (NotImplementedError, KeyError) as err:
                raise ParameterError(err)
        return self.__value

    @value.setter
    def value(self, value: Any):
        self.__clear_descendants()
        try:
            check_type("value", value, self.__type_info)
        except TypeError as err:
            raise ParameterError(
                "interface_function_handle annotation does not match given value"
            ) from err
        if self.__const and self.__value is not None:
            raise ParameterError("Cannot overwrite constant parameter.")
        self.__value = value

    def clear(self, clear_descendants=True):
        r"""Clear the value stored in this and dependent `Parameter`\ s."""
        if clear_descendants:
            self.__clear_descendants()
        if not self.__const:
            self.__value = None

    @property
    def parents(self) -> Dict[Enum, Parameter]:
        r"""Parent `Parameter`\ s of this `Parameter`\ s instance."""
        return self.__parents.copy()

    @property
    def children(self) -> Dict[Enum, Parameter]:
        r"""`Parameter`\ s with a child relation to this `Parameter`\ s instance."""
        return self.__children.copy()

    def add_child(self, child: Parameter):
        r"""Assign a Parameter a child relation"""
        if child.enum_id in self.__parents:
            raise ParameterError(
                "Neighbour can't be parent and child at the same time."
            )
        self.__children[child.enum_id] = child

    @property
    def descendants(self) -> Dict[Enum, Parameter]:
        r"""`Parameter`\ s with a descendant relation to this `Parameter`\ s instance."""
        return self.__descendants.copy()

    def add_descendant(self, descendant: Parameter):
        r"""Assign a Parameter a child relation"""
        if descendant.enum_id in self.__parents:
            raise ParameterError(
                f"Parameter {descendant.enum_id} can't be parent and descendant at the same time."
            )
        self.__descendants[descendant.enum_id] = descendant

    def __clear_descendants(self):
        for descendant_enum_id in self.__descendants:
            self.__descendants[descendant_enum_id].clear(clear_descendants=False)


class ImageParameter(Parameter):
    """`Parameter` with additional display functionality for stored image."""

    def __init__(
        self,
        enum_id: Enum,
        value_definition: Union[Type[Any], Callable[[Dict[Enum, Parameter]], Any]],
        parent_parameters: Optional[Set[Parameter]] = None,
    ):
        super().__init__(enum_id, value_definition, parent_parameters)
        self.__title = None
        self.__bounds: Optional[Tuple[int, int, int, int]] = None
        self.__bounds_lookup: Optional[Enum] = None
        self.__cmap = "gray"
        self.__aspect: Optional[float] = None

    def display(self, axsp: Any = None):
        """
        Display the stored image using matplotlib.

        Parameters
        ----------
        aspx : matplotlib.axes._subplots.AxesSubplot
            AxesSubplot on which to draw

        Notes
        -----
        Uses the `title`, `cmap` and `bounds` properties of the `ImageParameter` class.
        """
        if axsp is None:
            fig = plt.figure()
            axsp = fig.subplots()
        if self.__title is not None:
            axsp.set_title(self.__title)
        if self.bounds is None:
            axsp.imshow(self.value, cmap=self.__cmap, aspect=self.__aspect)
        else:
            (r_lower, r_upper, c_lower, c_upper) = self.bounds
            axsp.imshow(
                self.value[r_lower:r_upper, c_lower:c_upper],
                cmap=self.__cmap,
                aspect=self.__aspect,
            )

    @property
    def title(self) -> Optional[str]:
        """The image title used by the display function."""
        return self.__title

    @title.setter
    def title(self, title):
        self.__title = title

    @property
    def bounds(self) -> Optional[Tuple[int, int, int, int]]:
        """Cropping boundaries for the display function."""
        if self.__bounds is None and self.__bounds_lookup is None:
            return None
        if self.__bounds is None:
            return cast(
                Optional[Tuple[int, int, int, int]],
                self.parents[cast(Enum, self.__bounds_lookup)].value,
            )
        return self.__bounds

    @bounds.setter
    def bounds(self, bounds: Tuple[int, int, int, int]):
        self.__bounds = bounds

    @property
    def cmap(self) -> str:
        """The colormap used by matplotlib."""
        return self.__cmap

    @cmap.setter
    def cmap(self, cmap: str):
        self.__cmap = cmap

    @property
    def aspect(self) -> Optional[float]:
        """The aspect ratio used by matplotlib."""
        return self.__aspect

    @aspect.setter
    def aspect(self, aspect: float):
        self.__aspect = aspect

    def set_bounds_lookup(self, bounds_enum_id: Enum):
        """Reference a boundary Parameter using its `enum_id` as a lookup for cropping."""
        self.__bounds_lookup = bounds_enum_id



"""
Module implementing the parameter classes for the LogPolarSolver Model

Author: Fabian A. Preiss
"""

from enum import Enum
from typing import Dict, Hashable, Optional, cast, Tuple

import numpy as np
from imgreg.util.methods import ImageMethods
from imgreg.models.logpolar.enums import LogPolParams
from imgreg.util.params import ImageParameter, Parameter
from skimage.registration import phase_cross_correlation
from skimage.transform import AffineTransform

# Root Parameters


class RefImageParam(ImageParameter):
    """
    Reference image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, image: Optional[np.ndarray] = None):
        super().__init__(LogPolParams.REF_IMG, np.ndarray)
        if image is not None:
            self.value = image
        self.title = "Reference Image"


class ModImageParam(ImageParameter):
    """
    Modified image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, image: Optional[np.ndarray] = None):
        super().__init__(LogPolParams.MOD_IMG, np.ndarray)
        if image is not None:
            self.value = image
        self.title = "Modified Image"


class GaussDiffParam(Parameter):
    """
    Lower and upper kernel size of the DoG filter.

    .. _wgaussdiff:
        https://en.wikipedia.org/wiki/Difference_of_Gaussians

    Attributes
    ----------
    value : tuple[float, float]

    Notes
    -----
    Typical ratios for image enhancement are in the order of 1:4 and 1:5. [#f3]_
    A ratio of 1.6 approximates the Laplacian of Gaussian filter.

    References
    ----------
    .. [#f3] `Wikipedia, "Difference of Gaussians" <wgaussdiff_>`_
    """

    def __init__(self, gauss_diff: Tuple[float, float]):
        # TODO mypy throws: Type application has too many types (1 expected)
        # expects tuple[float]
        super().__init__(LogPolParams.GAUSS_DIFF, Tuple[float, float])
        self.value = gauss_diff


class WindowWeightParam(Parameter):
    """
    Weighting factor scaling beween windowed image and image, range of [0,1].

    Attributes
    ----------
    value : str
    """

    def __init__(self, window_weight: float):
        super().__init__(LogPolParams.WINDOW_WEIGHT, float)
        self.value = window_weight


class WindowTypeParam(Parameter):
    """
    Window type for FFT filter, see
    `scipy.signal.windows.get_window <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.get_window.html>`_
    for possible choices.

    Attributes
    ----------
    value : str
    """

    def __init__(self, window_type: str):
        super().__init__(LogPolParams.WINDOW_TYPE, str)
        self.value = window_type


class WindowRadiusExpParam(Parameter):
    """
    Window radius exponent, larger wrexp => faster computation.

    Notes
    -----
    If a value larger then 1 is used, this introduces a lowpass filter.

    Attributes
    ----------
    value : float >= 1
    """

    def __init__(self, w_r_exp: float):
        super().__init__(LogPolParams.WINDOW_RADIUS_EXP, float)
        self.value = w_r_exp


class UpsamplingParam(Parameter):
    """
    Upsampling factor. 1 => no upsampling, 20 => precision to 1/20 of a pixel.

    Attributes
    ----------
    value : int
    """

    def __init__(self, upsampling: int):
        super().__init__(LogPolParams.UPSAMPLING, int)
        self.value = upsampling


# Derived Parameters


# NOTE: design currently leads to a full clearing of derived properties although
# bounds is only an optical property and not necessary for the algorithm
class BoundsParam(Parameter):
    """
    Boundaries for image slicing.

    Attributes
    ----------
    value : tuple[int, int, int, int]
    """

    def __init__(
        self, parent_parameters, bounds: Optional[Tuple[int, int, int, int]] = None
    ):
        super().__init__(
            LogPolParams.BOUNDS,
            BoundsParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if bounds is not None:
            self.value = bounds

    @staticmethod
    def _value_function_handle(
        params: Dict[Enum, Parameter]
    ) -> Tuple[int, int, int, int]:
        image = params[LogPolParams.REF_IMG].value
        wrexp = params[LogPolParams.WINDOW_RADIUS_EXP].value
        warp_radius = ImageMethods.compute_warp_radius(min(image.shape), wrexp)
        center = np.array(image.shape) // 2
        r_lower, r_upper = (
            center[0] - warp_radius,
            center[0] + warp_radius,
        )
        c_lower, c_upper = (
            center[1] - warp_radius,
            center[1] + warp_radius,
        )
        return (r_lower, r_upper, c_lower, c_upper)


def _calculate_gaussdiff(image_type: Enum, params: Dict[Enum, Parameter]) -> np.ndarray:
    return ImageMethods.compute_dgfw(
        params[image_type].value,
        params[LogPolParams.GAUSS_DIFF].value,
        params[LogPolParams.WINDOW_WEIGHT].value,
        params[LogPolParams.WINDOW_TYPE].value,
    )


class GaussDiffWindowRefParam(ImageParameter):
    """
    Gaussian difference filtered and windowed reference input image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, gauss_diff_ref_img=None):
        super().__init__(
            LogPolParams.GAUSS_DIFF_REF_IMG,
            GaussDiffWindowRefParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Filtered Reference Image"
        if gauss_diff_ref_img is not None:
            self.value = gauss_diff_ref_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return _calculate_gaussdiff(LogPolParams.REF_IMG, params)


class GaussDiffWindowModParam(ImageParameter):
    """
    Gaussian difference filtered and windowed modified input image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, gauss_diff_mod_img=None):
        super().__init__(
            LogPolParams.GAUSS_DIFF_MOD_IMG,
            GaussDiffWindowModParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Filtered Modified Image"
        if gauss_diff_mod_img is not None:
            self.value = gauss_diff_mod_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return _calculate_gaussdiff(LogPolParams.MOD_IMG, params)


class FourierRefParam(ImageParameter):
    """
    Fourier transformed and filtered reference image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, fourier_ref_img=None):
        super().__init__(
            LogPolParams.FOURIER_REF_IMG,
            FourierRefParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Reference FFT Image\n(magnitude; zoomed)"
        self.cmap = "magma"
        if fourier_ref_img is not None:
            self.value = fourier_ref_img
        self.set_bounds_lookup(LogPolParams.BOUNDS)

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.compute_afts(params[LogPolParams.GAUSS_DIFF_REF_IMG].value)


class FourierModParam(ImageParameter):
    """
    Fourier transformed and filtered modified image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, fourier_mod_img=None):
        super().__init__(
            LogPolParams.FOURIER_MOD_IMG,
            FourierModParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Modified FFT Image\n(magnitude; zoomed)"
        self.cmap = "magma"
        if fourier_mod_img is not None:
            self.value = fourier_mod_img
        self.set_bounds_lookup(LogPolParams.BOUNDS)

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.compute_afts(params[LogPolParams.GAUSS_DIFF_MOD_IMG].value)


class WarpedFourierRefParam(ImageParameter):
    """
    Log-polar transformed of the fourier transformed reference image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, warped_fourier_ref_img=None):
        super().__init__(
            LogPolParams.WARPED_FOURIER_REF_IMG,
            WarpedFourierRefParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Reference FFT Image\n(Log-Polar-Transformed)"
        self.cmap = "magma"
        if warped_fourier_ref_img is not None:
            self.value = warped_fourier_ref_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.compute_log_polar_tf(
            params[LogPolParams.FOURIER_REF_IMG].value,
            params[LogPolParams.WINDOW_RADIUS_EXP].value,
        )


class WarpedFourierModParam(ImageParameter):
    """
    Log-polar transformed of the fourier transformed modified image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, warped_fourier_mod_img=None):
        super().__init__(
            LogPolParams.WARPED_FOURIER_MOD_IMG,
            WarpedFourierModParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Modified FFT Image\n(Log-Polar-Transformed)"
        self.cmap = "magma"
        if warped_fourier_mod_img is not None:
            self.value = warped_fourier_mod_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.compute_log_polar_tf(
            params[LogPolParams.FOURIER_MOD_IMG].value,
            params[LogPolParams.WINDOW_RADIUS_EXP].value,
        )


class RecoveredRotationScalePhaseParam(Parameter):
    """
    Recovered rotation angle in degrees and scaling factor including errors.

    Notes
    -----
    The errors are a lower estimate under ideal assumptions and can be much larger depending on the data.

    Attributes
    ----------
    value : tuple[np.ndarray, np.ndarray, dict[str, Hashable]]
    """

    def __init__(self, parent_parameters, recovered_rotation_scale_phase=None):
        super().__init__(
            LogPolParams.RECOVERED_ROTATION_SCALE_PHASE,
            RecoveredRotationScalePhaseParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_rotation_scale_phase is not None:
            self.value = recovered_rotation_scale_phase

    @staticmethod
    def _value_function_handle(
        params: Dict[Enum, Parameter]
    ) -> Tuple[np.ndarray, np.ndarray, Dict[str, Hashable]]:
        return ImageMethods.recover_rs(
            params[LogPolParams.WARPED_FOURIER_REF_IMG].value,
            params[LogPolParams.WARPED_FOURIER_MOD_IMG].value,
            params[LogPolParams.REF_IMG].value.shape,
            params[LogPolParams.UPSAMPLING].value,
            params[LogPolParams.WINDOW_RADIUS_EXP].value,
        )


class RecoveredRotationParam(Parameter):
    """
    Recovered rotation angle and error in degrees.

    Notes
    -----
    The errors are a lower estimate under ideal assumptions and can be much larger depending on the data.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rotation=None):
        super().__init__(
            LogPolParams.RECOVERED_ROTATION,
            RecoveredRotationParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_rotation is not None:
            self.value = recovered_rotation

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        rotation, _, _ = params[LogPolParams.RECOVERED_ROTATION_SCALE_PHASE].value
        return cast(np.ndarray, rotation)


class RecoveredScaleParam(Parameter):
    """
    Recovered scaling factor and error in degrees.

    Notes
    -----
    The errors are a lower estimate under ideal assumptions and can be much larger depending on the data.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_scale=None):
        super().__init__(
            LogPolParams.RECOVERED_SCALE,
            RecoveredScaleParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_scale is not None:
            self.value = recovered_scale

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        _, scale, _ = params[LogPolParams.RECOVERED_ROTATION_SCALE_PHASE].value
        return cast(np.ndarray, scale)


class RecoveredRotScaleParam(ImageParameter):
    """
    Rotation and scaling recovered image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rot_scale_img=None):
        super().__init__(
            LogPolParams.RECOVERED_ROT_SCALE_IMG,
            RecoveredRotScaleParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Rotation Recovered Image"
        if recovered_rot_scale_img is not None:
            self.value = recovered_rot_scale_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.compute_rts(
            params[LogPolParams.MOD_IMG].value,
            angle=params[LogPolParams.RECOVERED_ROTATION].value[0],
            scale=params[LogPolParams.RECOVERED_SCALE].value[0],
            translation=(0, 0),
            inverse=True,
            preserve_range=True,
        )


# TODO rename angle, rotation such that angle in deg, rot in rad
# TODO possibly define seperate function in ImageMethods
class RecoveredTranslationParam(Parameter):
    """
    Recovered x,y translation vector and error.

    Notes
    -----
    The errors are a lower estimate under ideal assumptions and can be much larger depending on the data.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_translation=None):
        super().__init__(
            LogPolParams.RECOVERED_TRANSLATION,
            RecoveredTranslationParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_translation is not None:
            self.value = recovered_translation

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        shifts, error, _ = phase_cross_correlation(
            params[LogPolParams.REF_IMG].value,
            params[LogPolParams.RECOVERED_ROT_SCALE_IMG].value,
            upsample_factor=params[LogPolParams.UPSAMPLING].value,
        )
        tform = AffineTransform(
            scale=params[LogPolParams.RECOVERED_SCALE].value[0],
            rotation=np.deg2rad(-params[LogPolParams.RECOVERED_ROTATION].value[0]),
        )
        shifts = tform.params[:2, :2] @ shifts[::-1]
        return np.array([*shifts, error])


class RecoveredRotScaleTr(ImageParameter):
    """
    Rotation, scaling and translation recovered image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rot_scale_tr_img=None):
        super().__init__(
            LogPolParams.RECOVERED_ROT_SCALE_TR_IMG,
            RecoveredRotScaleTr._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Recovered Image"
        if recovered_rot_scale_tr_img is not None:
            self.value = recovered_rot_scale_tr_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.compute_rts(
            params[LogPolParams.MOD_IMG].value,
            angle=params[LogPolParams.RECOVERED_ROTATION].value[0],
            scale=params[LogPolParams.RECOVERED_SCALE].value[0],
            translation=params[LogPolParams.RECOVERED_TRANSLATION].value[:-1],
            inverse=True,
        )



"""
A validator for image similarity.

Author: Fabian A. Preiss
"""

import numpy as np
from imgreg.models.validator.params import (
    ImageParam,
    RefImageParam,
    NormRelL2Param,
    AbsoluteDifferenceParam,
    SquaredDifferenceParam,
)
from imgreg.util.solver import Solver
from imgreg.models.validator.enums import ValidatorParams
from typing import Optional


class Validator(Solver):
    r"""
    Implements a validator model for image comparison.

    Parameters
    ----------
    img : numpy.ndarray
        The input image (one color channel only).
    ref_img : numpy.ndarray
        The reference image (one color channel only).

    Notes
    -----
    The model implements the following dependency graph to construct it's `Parameter`\ s.

        .. graphviz:: ../_assets/dependencies_validator.gv

    The `Parameter`\ s are documented in :ref:`imgreg.models.validator.params`.

    Examples
    --------
    We can visualize the internal `ImageParameter`\ s of the model as follows:

    .. plot:: tutorial/pyplots/validator.py
       :include-source:

    If we simply want to create a model and access the recovered values we first setup a model:

    >>> import numpy as np
    >>> import imgreg.data as data
    >>> from imgreg.models.validator import Validator
    >>> from imgreg.util.methods import ImageMethods
    >>> ref_img = np.array(data.ref_img())
    >>> img = ImageMethods.compute_rts(ref_img, angle=2, translation=(6,2))
    >>> val = Validator(img, ref_img)

    Calculate the relative norm of difference between the images.

    >>> val.NORM_REL_L2.value
    0.4875079942792268

    Note how this value approaches zero, as the image increase in their overlap:

    >>> val.IMG.value = ImageMethods.compute_rts(ref_img, angle=1, translation=(1,2))
    >>> val[ValidatorParams.NORM_REL_L2].value
    0.3942652180108632
    """

    def __init__(
        self, img: Optional[np.ndarray] = None, ref_img: Optional[np.ndarray] = None
    ):
        super().__init__()
        p_img = ImageParam(img)
        p_ref_img = RefImageParam(ref_img)
        p_norm_rel_l2 = NormRelL2Param({p_img, p_ref_img})
        p_absolute_difference = AbsoluteDifferenceParam({p_img, p_ref_img})
        p_squared_difference = SquaredDifferenceParam({p_img, p_ref_img})
        self._register_params(
            {
                p_img,
                p_ref_img,
                p_norm_rel_l2,
                p_absolute_difference,
                p_absolute_difference,
                p_squared_difference,
            }
        )
        self._generate_dep_graphs()
        for key in ValidatorParams:
            setattr(self, key.name, self[key])



"""
Solvermodel primitives.

Author: Fabian A. Preiss
"""
from __future__ import annotations
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Iterable,
    Sequence,
    Set,
    Union,
    cast,
)
from imgreg.util.params import Parameter, ImageParameter
from imgreg.util.graph import DAGraph
from enum import Enum
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from graphviz import Digraph


class SolverError(Exception):
    pass


def dependency_graph(parameters: Set[Parameter], invert=False) -> DAGraph:
    """Construct a `DAGraph` dependency graph given a `Parameter` set."""
    dep_graph = {}
    for parameter in parameters:
        if not invert:
            dep_graph[parameter] = {parent for parent in parameter.parents.values()}
        else:
            dep_graph[parameter] = {child for child in parameter.children.values()}
    return DAGraph(dep_graph)


class Solver:
    r"""
    Interface for a Solvermodel constructed from a set of parameters.

    Constructs a dependency graph for the `Parameter`\ s of the model and allows for
    lazy evaluation of the `Parameter`\ s.

    Notes
    -----
    The state of this class is cached, when a parameter is changed on which later states depend, the
    properties ascendant to said parameter are removed from the cache. `Solver._get_dep_graph()`
    allows access to the internal dependency graph representation.
    """

    def __init__(self, parameters: Optional[Set[Parameter]] = None):
        self.__params: Dict[Enum, Parameter] = dict()
        self.__dependency_graph: Optional[DAGraph] = None
        self.__idependency_graph: Optional[DAGraph] = None
        if parameters is not None:
            self._register_params(parameters)
            self._generate_dep_graphs()
            for param in parameters:
                setattr(self, param.enum_id.name, self[param.enum_id])

    # TODO in concrete solver unittest if all Parameter objects are registered
    def __getitem__(self, item: Enum) -> Parameter:
        return self.__params[item]

    def _register_params(self, params: Iterable[Parameter]) -> None:
        for param in params:
            self._register_param(param)

    def _register_param(self, param: Parameter) -> None:
        if param.enum_id in self.__params:
            raise SolverError(
                f"Cannot register parameter '{param.enum_id.value}' two times."
            )
        self.__params[param.enum_id] = param
        for parent in param.parents.values():
            parent.add_child(param)

    # TODO insert descendants into each parameter here, then replace add_child methods etc.
    def _generate_dep_graphs(self) -> None:
        self.__dependency_graph = dependency_graph(
            set(self.__params.values()), invert=False
        )
        self.__idependency_graph = dependency_graph(
            set(self.__params.values()), invert=True
        )
        for (
            parameter,
            ascendants,
        ) in self.__idependency_graph.vertex_ascendants_dict.items():
            for ascendant in ascendants:
                parameter.add_descendant(ascendant)

    def _get_dep_graph(self, invert=False) -> DAGraph:
        if self.__dependency_graph is None:
            self._generate_dep_graphs()

        result = self.__idependency_graph if invert else self.__dependency_graph
        return cast(DAGraph, result)

    def display(
        self,
        param_list: Sequence[Union[Enum, ImageParameter]],
        title: Optional[str] = None,
    ) -> None:
        """
        Fancy plot functionality for registered ImageParameters.

        Parameters
        ----------
        plotlist : sequence
            sequence, to access the ImageParameters registered in solver
        title : str
            str, contains title of overall plot
        """
        if len(param_list) == 0:
            return
        for param in param_list:
            if not isinstance(param, ImageParameter):
                try:
                    if not isinstance(self[param], ImageParameter):
                        raise SolverError(
                            f"{param} does not reference an ImageParameter."
                        )
                except KeyError as err:
                    raise SolverError from err
        n_rows = (1 + len(param_list)) // 2
        fig, _ = plt.subplots(n_rows, 2, figsize=(8, 4 * n_rows))
        gs = gridspec.GridSpec(n_rows, 4, fig)
        plt.subplots_adjust(wspace=0.4, hspace=0.3)
        for i in range(n_rows - 1 if len(param_list) % 2 else n_rows):
            ax_left = plt.subplot(gs[i : i + 1, :2])
            ax_right = plt.subplot(gs[i : i + 1, 2:])
            param = cast(
                ImageParameter,
                param_list[2 * i]
                if isinstance(param_list[2 * i], ImageParameter)
                else self[cast(Enum, param_list[2 * i])],
            )
            param.display(ax_left)
            param = cast(
                ImageParameter,
                param_list[2 * i + 1]
                if isinstance(param_list[2 * i + 1], ImageParameter)
                else self[cast(Enum, param_list[2 * i + 1])],
            )
            param.display(ax_right)
        if len(param_list) % 2:
            ax_center = plt.subplot(gs[-1, 1:3])
            param = cast(
                ImageParameter,
                param_list[-1]
                if isinstance(param_list[-1], ImageParameter)
                else self[param_list[-1]],
            )
            cast(ImageParameter, param).display(ax_center)
        fig.suptitle(title)
        plt.show()

    def dot_graph(
        self, node_args_func: Optional[Callable[[Parameter], Dict[str, str]]] = None
    ) -> Digraph:
        """Return a dot graph representation of the solver model."""
        vertex_parent_dict = self._get_dep_graph().vertex_parent_dict
        return vertex_parent_dict_to_dot(vertex_parent_dict, node_args_func)


def dot_args_box_img_func(parameter: Parameter) -> Dict[str, str]:
    """
    Generate the shape argument for a dot graph depending on the given Parameter

    Parameters
    ----------
    parameter : Parameter
        The parameter of a node
    """
    return {"shape": "box" if isinstance(parameter, ImageParameter) else "oval"}


def vertex_parent_dict_to_dot(
    vertex_parent_dict: Dict[Parameter, Set[Parameter]],
    node_args_func: Optional[Callable[[Parameter], Dict[str, str]]] = None,
    invert=False,
) -> Digraph:
    """
    Convert a directed graph to a dot graph.

    Parameters
    ----------
    vertex_parent_dict : dict[Hashable, set[Hashable]]
        A dictionary representation of the directed graph
    node_args_func : Callable[[Hashable], dict[str, str]]
        A function handle to generate keyword arguments for the node of the dot graph
        depending on the current vertex
    """
    node_args_func = (lambda x: {}) if node_args_func is None else node_args_func
    dot = Digraph(comment="dependencies")
    for parameter in vertex_parent_dict:
        dot.node(parameter.enum_id.name, **node_args_func(parameter))
    for node, deps in vertex_parent_dict.items():
        for dep in deps:
            if invert:
                dot.edge(node.enum_id.name, dep.enum_id.name)
            else:
                dot.edge(dep.enum_id.name, node.enum_id.name)
    return dot



"""
Module implementing the parameter classes for the angle selector Model

Author: Fabian A. Preiss
"""
from enum import Enum
from typing import Dict, Optional, cast

import numpy as np
from skimage.registration import phase_cross_correlation
from skimage.transform import AffineTransform

from imgreg.models.angleselect.enums import AngleSelectParams
from imgreg.util.methods import ImageMethods
from imgreg.util.params import ImageParameter, Parameter


class ImageParam(ImageParameter):
    """
    Image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, image: Optional[np.ndarray] = None):
        super().__init__(AngleSelectParams.IMG, np.ndarray)
        if image is not None:
            self.value = image
        self.title = "Image"


class RefImageParam(ImageParameter):
    """
    Reference image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, image: Optional[np.ndarray] = None):
        super().__init__(AngleSelectParams.REF_IMG, np.ndarray)
        if image is not None:
            self.value = image
        self.title = "Reference Image"


class AngleAParam(Parameter):
    """
    Candidate Angle A.

    Attributes
    ----------
    value : float
    """

    def __init__(self, angle_a: float):
        super().__init__(AngleSelectParams.ANGLE_A, float)
        self.value = angle_a


class AngleBParam(Parameter):
    """
    Candidate Angle A.

    Attributes
    ----------
    value : float
    """

    def __init__(self, angle_b: float):
        super().__init__(AngleSelectParams.ANGLE_B, float)
        self.value = angle_b


class UpsamplingParam(Parameter):
    """
    Upsampling factor. 1 => no upsampling, 20 => precision to 1/20 of a pixel.

    Attributes
    ----------
    value : int
    """

    def __init__(self, upsampling: int):
        super().__init__(AngleSelectParams.UPSAMPLING, int)
        self.value = upsampling


# derived properties


class RotAParam(ImageParameter):
    """
    Image Rotated A.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rot_img=None):
        super().__init__(
            AngleSelectParams.ROT_A_IMG,
            RotAParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Image Rotated A"
        if recovered_rot_img is not None:
            self.value = recovered_rot_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.compute_rts(
            params[AngleSelectParams.IMG].value,
            angle=params[AngleSelectParams.ANGLE_A].value,
            scale=1,
            translation=(0, 0),
            inverse=True,
            preserve_range=True,
        )


class RotBParam(ImageParameter):
    """
    Image Rotated B.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rot_img=None):
        super().__init__(
            AngleSelectParams.ROT_B_IMG,
            RotBParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Image Rotated B"
        if recovered_rot_img is not None:
            self.value = recovered_rot_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.compute_rts(
            params[AngleSelectParams.IMG].value,
            angle=params[AngleSelectParams.ANGLE_B].value,
            scale=1,
            translation=(0, 0),
            inverse=True,
            preserve_range=True,
        )


class TranslationAParam(Parameter):
    """
    Recovered x,y translation vector and error for A.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_translation=None):
        super().__init__(
            AngleSelectParams.TRANSLATION_A,
            TranslationAParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_translation is not None:
            self.value = recovered_translation

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        shifts, error, _ = phase_cross_correlation(
            params[AngleSelectParams.REF_IMG].value,
            params[AngleSelectParams.ROT_A_IMG].value,
            upsample_factor=params[AngleSelectParams.UPSAMPLING].value,
        )
        tform = AffineTransform(
            scale=1,
            rotation=np.deg2rad(-params[AngleSelectParams.ANGLE_A].value),
        )
        shifts = tform.params[:2, :2] @ shifts[::-1]
        return np.array([*shifts, error])


class TranslationBParam(Parameter):
    """
    Recovered x,y translation vector and error for B.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_translation=None):
        super().__init__(
            AngleSelectParams.TRANSLATION_B,
            TranslationBParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_translation is not None:
            self.value = recovered_translation

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        shifts, error, _ = phase_cross_correlation(
            params[AngleSelectParams.REF_IMG].value,
            params[AngleSelectParams.ROT_B_IMG].value,
            upsample_factor=params[AngleSelectParams.UPSAMPLING].value,
        )
        tform = AffineTransform(
            scale=1,
            rotation=np.deg2rad(-params[AngleSelectParams.ANGLE_B].value),
        )
        shifts = tform.params[:2, :2] @ shifts[::-1]
        return np.array([*shifts, error])


class RotTrAParam(ImageParameter):
    """
    Rotation, and translation recovered image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rot_tr_img=None):
        super().__init__(
            AngleSelectParams.ROT_TR_A_IMG,
            RotTrAParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Recovered Image"
        if recovered_rot_tr_img is not None:
            self.value = recovered_rot_tr_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.compute_rts(
            params[AngleSelectParams.IMG].value,
            angle=params[AngleSelectParams.ANGLE_A].value,
            scale=1,
            translation=params[AngleSelectParams.TRANSLATION_A].value[:-1],
            inverse=True,
            preserve_range=True,
        )


class RotTrBParam(ImageParameter):
    """
    Rotation, and translation recovered image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rot_tr_img=None):
        super().__init__(
            AngleSelectParams.ROT_TR_B_IMG,
            RotTrBParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Recovered Image"
        if recovered_rot_tr_img is not None:
            self.value = recovered_rot_tr_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.compute_rts(
            params[AngleSelectParams.IMG].value,
            angle=params[AngleSelectParams.ANGLE_B].value,
            scale=1,
            translation=params[AngleSelectParams.TRANSLATION_B].value[:-1],
            inverse=True,
            preserve_range=True,
        )


class SelectorParam(Parameter):
    """
    Return True, if A matches better than B

    Attributes
    ----------
    value : bool
    """

    def __init__(self, parent_parameters, selection=None):
        super().__init__(
            AngleSelectParams.SELECTOR,
            SelectorParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if selection is not None:
            self.value = selection

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> bool:
        return ImageMethods.norm_rel_l2(
            params[AngleSelectParams.ROT_TR_A_IMG].value,
            params[AngleSelectParams.REF_IMG].value,
        ) <= ImageMethods.norm_rel_l2(
            params[AngleSelectParams.ROT_TR_B_IMG].value,
            params[AngleSelectParams.REF_IMG].value,
        )


class RecoveredRotationParam(Parameter):
    """
    The recovered rotation angle and error between the modified and reference image.

    Notes
    -----
    The errors are a lower estimate under ideal assumptions and can be much larger depending on the data.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rotation=None):
        super().__init__(
            AngleSelectParams.RECOVERED_ROTATION,
            RecoveredRotationParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_rotation is not None:
            self.value = recovered_rotation

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> float:
        return cast(
            float,
            params[AngleSelectParams.ANGLE_A].value
            if params[AngleSelectParams.SELECTOR].value
            else params[AngleSelectParams.ANGLE_B].value,
        )


class RecoveredRotParam(ImageParameter):
    """
    Rotation recovered image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rot_img=None):
        super().__init__(
            AngleSelectParams.RECOVERED_ROT_IMG,
            RecoveredRotParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_rot_img is not None:
            self.value = recovered_rot_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return cast(
            np.ndarray,
            params[AngleSelectParams.ROT_A_IMG].value
            if params[AngleSelectParams.SELECTOR].value
            else params[AngleSelectParams.ROT_B_IMG].value,
        )


class RecoveredTranslationParam(Parameter):
    """
    Recovered x,y translation vector and error.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_translation=None):
        super().__init__(
            AngleSelectParams.RECOVERED_TRANSLATION,
            RecoveredTranslationParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_translation is not None:
            self.value = recovered_translation

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return cast(
            np.ndarray,
            params[AngleSelectParams.TRANSLATION_A].value
            if params[AngleSelectParams.SELECTOR].value
            else params[AngleSelectParams.TRANSLATION_B].value,
        )


class RecoveredRotTrParam(ImageParameter):
    """
    Rotation, and translation recovered image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rot_tr_img=None):
        super().__init__(
            AngleSelectParams.RECOVERED_ROT_TR_IMG,
            RecoveredRotTrParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_rot_tr_img is not None:
            self.value = recovered_rot_tr_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return cast(
            np.ndarray,
            params[AngleSelectParams.ROT_TR_A_IMG].value
            if params[AngleSelectParams.SELECTOR].value
            else params[AngleSelectParams.ROT_TR_B_IMG].value,
        )



"""
A solver for angle selection

Author: Fabian A. Preiss
"""

import numpy as np

from imgreg.models.angleselect.enums import AngleSelectParams
from imgreg.models.angleselect.params import (
    AngleAParam,
    AngleBParam,
    ImageParam,
    RecoveredRotationParam,
    RecoveredRotParam,
    RecoveredRotTrParam,
    RecoveredTranslationParam,
    RefImageParam,
    RotAParam,
    RotBParam,
    RotTrAParam,
    RotTrBParam,
    SelectorParam,
    TranslationAParam,
    TranslationBParam,
    UpsamplingParam,
)
from imgreg.util.solver import Solver


class AngleSelect(Solver):
    r"""
    Implements a model for angle selection

    Parameters
    ----------
    img : numpy.ndarray
        The input image (one color channel only).
    ref_img : numpy.ndarray
        The reference image (one color channel only).
    angle_a : float
        Angle choice a in degrees
    angle_b : float
        Angle choice b in degrees
    upsampling : int
        Upsampling factor. 1 => no upsampling, 20 => precision to 1/20 of a pixel.

    Notes
    -----

    The model implements the following dependency graph to construct it's `Parameter`\ s.

        .. graphviz:: ../_assets/dependencies_angleselect.gv
    """

    def __init__(
        self,
        img: np.ndarray,
        ref_img: np.ndarray,
        angle_a: float,
        angle_b: float,
        upsampling: int = 10,
    ):
        super().__init__()
        p_img = ImageParam(img)
        p_ref_img = RefImageParam(ref_img)
        p_angle_a = AngleAParam(angle_a)
        p_angle_b = AngleBParam(angle_b)
        p_upsampling = UpsamplingParam(upsampling)
        p_rot_a_img = RotAParam({p_img, p_angle_a})
        p_rot_b_img = RotBParam({p_img, p_angle_b})
        p_translation_a = TranslationAParam(
            {p_ref_img, p_rot_a_img, p_upsampling, p_angle_a}
        )
        p_translation_b = TranslationBParam(
            {p_ref_img, p_rot_b_img, p_upsampling, p_angle_b}
        )
        p_rot_tr_a_img = RotTrAParam({p_img, p_angle_a, p_translation_a})
        p_rot_tr_b_img = RotTrBParam({p_img, p_angle_b, p_translation_b})
        p_selector = SelectorParam({p_rot_tr_a_img, p_rot_tr_b_img, p_ref_img})
        p_recovered_rotation = RecoveredRotationParam(
            {p_selector, p_angle_a, p_angle_b}
        )
        p_recovered_rot_img = RecoveredRotParam({p_selector, p_rot_a_img, p_rot_b_img})
        p_recovered_translation = RecoveredTranslationParam(
            {p_selector, p_translation_a, p_translation_b}
        )
        p_recovered_rot_tr_img = RecoveredRotTrParam(
            {p_selector, p_rot_tr_a_img, p_rot_tr_b_img}
        )
        self._register_params(
            {
                p_img,
                p_ref_img,
                p_angle_a,
                p_angle_b,
                p_upsampling,
                p_rot_a_img,
                p_rot_b_img,
                p_translation_a,
                p_translation_b,
                p_rot_tr_a_img,
                p_rot_tr_b_img,
                p_selector,
                p_recovered_rotation,
                p_recovered_rot_img,
                p_recovered_translation,
                p_recovered_rot_tr_img,
            }
        )
        self._generate_dep_graphs()
        for key in AngleSelectParams:
            setattr(self, key.name, self[key])



"""
Directory processing utilities

Author: Fabian A. Preiss
"""
import os
import fnmatch
from typing import Union, List, Set


def fnmatch_filter(
    ls_files: Set[str], pattern: Union[str, List[str]] = "*"
) -> Set[str]:
    """Return the subset of strings that match given patterns."""
    if isinstance(pattern, str):
        patterns = [pattern]
    return {
        fname
        for fname in ls_files
        if True in [fnmatch.fnmatch(fname, p) for p in patterns]
    }


class DirectoryViewError(Exception):
    pass


class DirectoryView:
    def __init__(self, inputdir=".", file_pattern="*"):
        inputdir = inputdir[:-1] if inputdir[-1] == "/" else inputdir
        inputdir = os.path.expanduser(inputdir)
        if not os.path.isdir(inputdir):
            raise DirectoryViewError(f"{inputdir} is not a directory")
        ls_current_dir = os.listdir(inputdir)
        self.realpath = os.path.realpath(inputdir)
        self.basename = os.path.basename(self.realpath)
        self.files = fnmatch_filter(
            {
                file_name
                for file_name in ls_current_dir
                if os.path.isfile(inputdir + "/" + file_name)
            },
            file_pattern,
        )
        self.dirs = {
            DirectoryView(self.realpath + "/" + d, file_pattern=file_pattern)
            for d in ls_current_dir
            if os.path.isdir(self.realpath + "/" + d)
        }

    def file_path_generator(self, step: int = 1):
        for i, file in enumerate(sorted(self.files)):
            if not i % step:
                yield f"{self.realpath}/{file}"


# TODO: make it possible to give a filetree
# dict and ignore folders not in filetree
def dirtree_reader(inputdir=".", file_pattern="*"):
    inputdir = inputdir[:-1] if inputdir[-1] == "/" else inputdir
    if not os.path.isdir(inputdir):
        print(f"{inputdir} is not a directory")
        return
    ls_current_dir = os.listdir(inputdir)
    ls_files = {x for x in ls_current_dir if os.path.isfile(inputdir + "/" + x)}
    ls_files = fnmatch_filter(ls_files, file_pattern)
    ls_dirs = [x for x in ls_current_dir if os.path.isdir(inputdir + "/" + x)]
    dict_dirs = [
        dirtree_reader(inputdir + "/" + x, file_pattern=file_pattern) for x in ls_dirs
    ]
    dict_dirs = {} if dict_dirs == [] else {"dirs": dict_dirs}
    dict_files = {} if ls_files == set() else {"files": ls_files}
    dict_dirs.update(dict_files)
    return {os.path.basename(inputdir): dict_dirs}


class File_Set_Ops:
    @staticmethod
    def a_or_b(a, b):
        return {"files": a["files"].union(b["files"])}

    @staticmethod
    def a_and_b(a, b):
        return {"files": a["files"].intersection(b["files"])}

    @staticmethod
    def a_diff_b(a, b):
        return {"files": a["files"].difference(b["files"])}

    @staticmethod
    def a_xor_b(a, b):
        return {"files": a["files"].symmetric_difference(b["files"])}

    @staticmethod
    def b_diff_a(a, b):
        return {"files": b["files"].difference(a["files"])}

    @staticmethod
    def a_b_op(a, b, op="or"):
        methods = {
            "or": File_Set_Ops.a_or_b,
            "and": File_Set_Ops.a_and_b,
            "\\": File_Set_Ops.a_diff_b,
            "xor": File_Set_Ops.a_xor_b,
            "/": File_Set_Ops.b_diff_a,
        }
        return methods[op](a, b)

    @staticmethod
    def a_copy(a):
        return {"files": a["files"].copy()}

    @staticmethod
    def a_set(a):
        return set()

    @staticmethod
    def a_op(a, op="or"):
        methods = {
            "or": File_Set_Ops.a_copy,
            "and": File_Set_Ops.a_set,
            "\\": File_Set_Ops.a_copy,
            "xor": File_Set_Ops.a_copy,
            "/": File_Set_Ops.a_set,
        }
        return methods[op](a)

    @staticmethod
    def file_set_op(a, b=set(), op="or"):
        if "files" in a and "files" in b:
            return File_Set_Ops.a_b_op(a, b, op)
        elif "files" in a and "files" not in b:
            return File_Set_Ops.a_op(a, op)
        elif "files" in b and "files" not in a:
            return File_Set_Ops.a_op(b, op)
        else:
            return {}



"""
Image registration based on the log-polar transform.

Author: Fabian A. Preiss
"""

import numpy as np
from imgreg.models.logpolar.params import (
    BoundsParam,
    FourierModParam,
    FourierRefParam,
    GaussDiffParam,
    GaussDiffWindowModParam,
    GaussDiffWindowRefParam,
    ModImageParam,
    RecoveredRotationParam,
    RecoveredRotationScalePhaseParam,
    RecoveredRotScaleParam,
    RecoveredRotScaleTr,
    RecoveredScaleParam,
    RecoveredTranslationParam,
    RefImageParam,
    UpsamplingParam,
    WarpedFourierModParam,
    WarpedFourierRefParam,
    WindowRadiusExpParam,
    WindowTypeParam,
    WindowWeightParam,
)
from imgreg.util.solver import Solver
from imgreg.models.logpolar.enums import LogPolParams
from typing import Optional

from PIL import Image


# NOTE: try applying exponential filter before Gaussdiff
class LogPolarSolver(Solver):
    r"""
    Implements an image registration model based on the log-polar transform.

    The model tries to reconstruct the difference of scale, rotation and
    translation between two images.

    Parameters
    ----------
    ref_img : numpy.ndarray
        The original input image (one color channel only).
    mod_img : numpy.ndarray
        The modified input image (one color channel only).

    Notes
    -----
    Build based on the approach of the example code [#f1]_ from scikit-image. Alternative
    implementations using feature based detection algorithms are shown in [#f2]_.

    The model implements the following dependency graph to construct it's `Parameter`\ s.

        .. graphviz:: ../_assets/dependencies_logpolar.gv

    The `Parameter`\ s are documented in :ref:`imgreg.models.logpolar.params`.

    References
    ----------
    .. [#f1] `Using Polar and Log-Polar Transformations for Registration
       <https://scikit-image.org/docs/stable/auto_examples/registration/plot_register_rotation.html#sphx-glr-auto-examples-registration-plot-register-rotation-py>`_
    .. [#f2] `ORB feature detector and binary descriptor
       <https://scikit-image.org/docs/dev/auto_examples/features_detection/plot_orb.html#sphx-glr-auto-examples-features-detection-plot-orb-py>`_

    Examples
    --------
    We can visualize the internal `ImageParameter`\ s of the model as follows:

    .. plot:: tutorial/pyplots/logpolarsolver.py
       :include-source:

    If we simply want to create a model and access the recovered values we first setup a model:

    >>> import numpy as np
    >>> import imgreg.data as data
    >>> from imgreg.models.logpolar import LogPolarSolver, LogPolParams
    >>> ref_img = np.array(data.ref_img())
    >>> mod_img = np.array(data.mod_img())
    >>> lps = LogPolarSolver(ref_img, mod_img)

    Now the parameters of the model can now be accessed as follows:

    >>> lps.RECOVERED_ROTATION.value
    array([-13.06730769,   0.11259774])

    >>> lps.RECOVERED_TRANSLATION.value
    array([-17.98318062,  31.037803  ,   0.42407651])
    """

    def __init__(
        self, ref_img: Optional[np.ndarray] = None, mod_img: Optional[np.ndarray] = None
    ):
        super().__init__()

        p_ref_img = RefImageParam(ref_img)
        p_mod_img = ModImageParam(mod_img)
        p_gauss_diff = GaussDiffParam((5, 20))
        p_window_weight = WindowWeightParam(1)
        p_window_type = WindowTypeParam("hann")
        p_window_radius_exp = WindowRadiusExpParam(3)
        p_upsampling = UpsamplingParam(10)

        p_bounds = BoundsParam({p_ref_img, p_window_radius_exp})
        p_gauss_diff_ref_img = GaussDiffWindowRefParam(
            {p_ref_img, p_gauss_diff, p_window_weight, p_window_type}
        )
        p_gauss_diff_mod_img = GaussDiffWindowModParam(
            {p_mod_img, p_gauss_diff, p_window_weight, p_window_type}
        )
        p_fourier_ref_img = FourierRefParam({p_gauss_diff_ref_img, p_bounds})
        p_fourier_mod_img = FourierModParam({p_gauss_diff_mod_img, p_bounds})

        p_warped_fourier_ref_img = WarpedFourierRefParam(
            {p_fourier_ref_img, p_window_radius_exp}
        )
        p_warped_fourier_mod_img = WarpedFourierModParam(
            {p_fourier_mod_img, p_window_radius_exp}
        )
        p_recovered_rotation_scale_phase = RecoveredRotationScalePhaseParam(
            {
                p_warped_fourier_ref_img,
                p_warped_fourier_mod_img,
                p_ref_img,
                p_upsampling,
                p_window_radius_exp,
            }
        )
        p_recovered_rotation = RecoveredRotationParam(
            {p_recovered_rotation_scale_phase}
        )
        p_recovered_scale = RecoveredScaleParam({p_recovered_rotation_scale_phase})
        p_recovered_rot_scale_img = RecoveredRotScaleParam(
            {p_mod_img, p_recovered_rotation, p_recovered_scale}
        )
        p_recovered_translation = RecoveredTranslationParam(
            {
                p_ref_img,
                p_recovered_rot_scale_img,
                p_upsampling,
                p_recovered_rotation,
                p_recovered_scale,
            }
        )
        p_recovered_rot_scale_tr = RecoveredRotScaleTr(
            {
                p_mod_img,
                p_recovered_rotation,
                p_recovered_scale,
                p_recovered_translation,
            }
        )
        self._register_params(
            {
                p_ref_img,
                p_mod_img,
                p_gauss_diff,
                p_window_weight,
                p_window_type,
                p_window_radius_exp,
                p_upsampling,
                p_bounds,
                p_gauss_diff_ref_img,
                p_gauss_diff_mod_img,
                p_fourier_ref_img,
                p_fourier_mod_img,
                p_warped_fourier_ref_img,
                p_warped_fourier_mod_img,
                p_recovered_rotation_scale_phase,
                p_recovered_rotation,
                p_recovered_scale,
                p_recovered_rot_scale_img,
                p_recovered_translation,
                p_recovered_rot_scale_tr,
            }
        )
        self._generate_dep_graphs()
        for key in LogPolParams:
            setattr(self, key.name, self[key])



"""
Module implementing the parameter classes for a validator.

Author: Fabian A. Preiss
"""

from enum import Enum
from typing import Dict, Optional

import numpy as np
from imgreg.util.methods import ImageMethods
from imgreg.models.validator.enums import ValidatorParams
from imgreg.util.params import ImageParameter, Parameter

# Root Parameters


class ImageParam(ImageParameter):
    """
    Image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, image: Optional[np.ndarray] = None):
        super().__init__(ValidatorParams.IMG, np.ndarray)
        if image is not None:
            self.value = image
        self.title = "Image"


class RefImageParam(ImageParameter):
    """
    Reference image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, image: Optional[np.ndarray] = None):
        super().__init__(ValidatorParams.REF_IMG, np.ndarray)
        if image is not None:
            self.value = image
        self.title = "Reference Image"


# derived properties


class NormRelL2Param(Parameter):
    """
    Relative L2 Norm similarity measurement for the images.

    Attributes
    ----------
    value : float
    """

    def __init__(self, parent_parameters):
        super().__init__(
            ValidatorParams.NORM_REL_L2,
            NormRelL2Param._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "norm(ref_image-image)/norm(ref_image)"

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> float:
        return ImageMethods.norm_rel_l2(
            params[ValidatorParams.IMG].value,
            params[ValidatorParams.REF_IMG].value,
        )


class AbsoluteDifferenceParam(ImageParameter):
    """
    Absolute value of the difference between the images.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters):
        super().__init__(
            ValidatorParams.ABSOLUTE_DIFFERENCE_IMG,
            AbsoluteDifferenceParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "abs(ref_image-image)"

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.abs_diff(
            params[ValidatorParams.IMG].value,
            params[ValidatorParams.REF_IMG].value,
        )


class SquaredDifferenceParam(ImageParameter):
    """
    Squared difference between the images.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters):
        super().__init__(
            ValidatorParams.SQUARED_DIFFERENCE_IMG,
            SquaredDifferenceParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "(ref_image-image)^2"

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return ImageMethods.sqr_diff(
            params[ValidatorParams.IMG].value,
            params[ValidatorParams.REF_IMG].value,
        )



from setuptools import setup

with open("README.md", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="imgreg",
    version="1.0.2",
    description="Image registration models with extendable interfaces.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Fabian A. Preiss, Jendrik A. Potyka",
    author_email="devops@digon.io",
    license="GPLv3",
    packages=[
        "imgreg",
        "imgreg.data",
        "imgreg.models",
        "imgreg.util",
        "imgreg.models.logpolar",
        "imgreg.models.angleselect",
        "imgreg.models.radon",
        "imgreg.models.validator",
    ],
    keywords="image registration data analysis science",
    package_data={"imgreg": ["data/ref_img.jpg", "data/mod_img.jpg"]},
    include_package_data=True,
    install_requires=[
        "numpy==1.20.3",
        "scipy==1.6.3",
        "scikit-image==0.18.1",
        "matplotlib==3.4.1",
        "typeguard==2.12.0",
        "graphviz==0.16",
        "Pillow==8.2.0",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries",
        "Operating System :: Unix",
        "Operating System :: POSIX",
    ],
    zip_safe=False,
    url="https://gitlab.com/DigonIO/imgreg",
)



"""
Module implementing the parameter classes for the radon Model

Author: Fabian A. Preiss
"""

from enum import Enum
from typing import Dict, Optional, cast

import numpy as np
from imgreg.util.methods import ImageMethods
from imgreg.models.radon.enums import RadonParams
from imgreg.util.params import ImageParameter, Parameter
from skimage.registration import phase_cross_correlation
from imgreg.models.angleselect import AngleSelect, AngleSelectParams

# Root Parameters


class RefImageParam(ImageParameter):
    """
    Reference image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, image: Optional[np.ndarray] = None):
        super().__init__(RadonParams.REF_IMG, np.ndarray)
        if image is not None:
            self.value = image
        self.title = "Reference Image"


class ModImageParam(ImageParameter):
    """
    Modified image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, image: Optional[np.ndarray] = None):
        super().__init__(RadonParams.MOD_IMG, np.ndarray)
        if image is not None:
            self.value = image
        self.title = "Modified Image"


class AngularPrecisionParam(Parameter):
    """
    Targeted angular precision in degrees.

    Notes
    -----
    The targeted error is the lower estimate under ideal assumptions, in practise won't be achieavable.

    Attributes
    ----------
    value : float
    """

    def __init__(self, angular_precision: float):
        super().__init__(RadonParams.ANGULAR_PRECISION, float)
        self.value = angular_precision


class ExponentialFilterSignalNoiseParam(Parameter):
    """
    Signal to noise ratio when applying the min-max exponential filter.

    Attributes
    ----------
    value : float
    """

    def __init__(self, exponential_filter_signal_noise: float):
        super().__init__(RadonParams.EXPONENTIAL_FILTER_SIGNAL_NOISE, float)
        self.value = exponential_filter_signal_noise


class ThetaParam(Parameter):
    """
    The initial search angles.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, theta: np.ndarray):
        super().__init__(RadonParams.THETA, np.ndarray)
        self.value = theta


class UpsamplingParam(Parameter):
    """
    Upsampling factor. 1 => no upsampling, 20 => precision to 1/20 of a pixel.

    Attributes
    ----------
    value : int
    """

    def __init__(self, upsampling: int):
        super().__init__(RadonParams.UPSAMPLING, int)
        self.value = upsampling


# derived properties


class RefRotationParam(Parameter):
    """
    Recovered rotation angle and error in degrees of the reference image.

    Notes
    -----
    The errors are a lower estimate under ideal assumptions and can be much larger depending on the data.

    Attributes
    ----------
    value : float
    """

    def __init__(self, parent_parameters, ref_rotation: Optional[float] = None):
        super().__init__(
            RadonParams.REF_ROTATION,
            RefRotationParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if ref_rotation is not None:
            self.value = ref_rotation

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return np.array(
            [
                ImageMethods.max_sinogram_angle(
                    params[RadonParams.REF_IMG].value,
                    theta=params[RadonParams.THETA].value,
                    exp_filter_val=params[
                        RadonParams.EXPONENTIAL_FILTER_SIGNAL_NOISE
                    ].value,
                    precision=params[RadonParams.ANGULAR_PRECISION].value,
                ),
                params[RadonParams.ANGULAR_PRECISION].value,
            ]
        )


class ModRotationParam(Parameter):
    """
    Recovered rotation angle and error in degrees of the modified image.

    Notes
    -----
    The errors are a lower estimate under ideal assumptions and can be much larger depending on the data.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, mod_rotation: Optional[float] = None):
        super().__init__(
            RadonParams.MOD_ROTATION,
            ModRotationParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if mod_rotation is not None:
            self.value = mod_rotation

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return np.array(
            [
                ImageMethods.max_sinogram_angle(
                    params[RadonParams.MOD_IMG].value,
                    theta=params[RadonParams.THETA].value,
                    exp_filter_val=params[
                        RadonParams.EXPONENTIAL_FILTER_SIGNAL_NOISE
                    ].value,
                    precision=params[RadonParams.ANGULAR_PRECISION].value,
                ),
                params[RadonParams.ANGULAR_PRECISION].value,
            ]
        )


class RotationCandidateParam(Parameter):
    """
    Candidate for the recovered rotation angle + error between the modified and reference image.

    Notes
    -----
    The errors are a lower estimate under ideal assumptions and can be much larger depending on the data.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(
        self, parent_parameters, rotation_candidate: Optional[np.ndarray] = None
    ):
        super().__init__(
            RadonParams.ROTATION_CANDIDATE,
            RotationCandidateParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if rotation_candidate is not None:
            self.value = rotation_candidate

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        return np.array(
            [
                params[RadonParams.MOD_ROTATION].value[0]
                - params[RadonParams.REF_ROTATION].value[0],
                (
                    params[RadonParams.MOD_ROTATION].value[1] ** 2
                    + params[RadonParams.REF_ROTATION].value[1] ** 2
                )
                ** 0.5,
            ]
        )


class AngleSelectParam(Parameter):
    """A solver model for angle selection"""

    def __init__(self, parent_parameters, angle_select: Optional[AngleSelect] = None):
        super().__init__(
            RadonParams.ANGLE_SELECT,
            AngleSelectParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if angle_select is not None:
            self.value = angle_select

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> AngleSelect:
        return AngleSelect(
            params[RadonParams.MOD_IMG].value,
            params[RadonParams.REF_IMG].value,
            params[RadonParams.ROTATION_CANDIDATE].value[0],
            params[RadonParams.ROTATION_CANDIDATE].value[0] + 180,
            params[RadonParams.UPSAMPLING].value,
        )


class RecoveredRotationParam(Parameter):
    """
    The recovered rotation angle and error between the modified and reference image.

    Notes
    -----
    The errors are a lower estimate under ideal assumptions and can be much larger depending on the data.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, angle_select: Optional[np.ndarray] = None):
        super().__init__(
            RadonParams.RECOVERED_ROTATION,
            RecoveredRotationParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if angle_select is not None:
            self.value = angle_select

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        matching_angle = params[RadonParams.ROTATION_CANDIDATE].value.copy()
        angle_select = params[RadonParams.ANGLE_SELECT].value
        matching_angle[0] = angle_select[AngleSelectParams.RECOVERED_ROTATION].value
        return cast(np.ndarray, matching_angle)


class RecoveredRotParam(ImageParameter):
    """
    Rotation recovered image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rot_img=None):
        super().__init__(
            RadonParams.RECOVERED_ROT_IMG,
            RecoveredRotParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Rotation Recovered Image"
        if recovered_rot_img is not None:
            self.value = recovered_rot_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        angle_select = params[RadonParams.ANGLE_SELECT].value
        return cast(np.ndarray, angle_select[AngleSelectParams.RECOVERED_ROT_IMG].value)


class RecoveredTranslationParam(Parameter):
    """
    Recovered x,y translation vector and error.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_translation=None):
        super().__init__(
            RadonParams.RECOVERED_TRANSLATION,
            RecoveredTranslationParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        if recovered_translation is not None:
            self.value = recovered_translation

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        angle_select = params[RadonParams.ANGLE_SELECT].value
        return cast(
            np.ndarray, angle_select[AngleSelectParams.RECOVERED_TRANSLATION].value
        )


class RecoveredRotTrParam(ImageParameter):
    """
    Rotation, and translation recovered image.

    Attributes
    ----------
    value : numpy.ndarray
    """

    def __init__(self, parent_parameters, recovered_rot_tr_img=None):
        super().__init__(
            RadonParams.RECOVERED_ROT_TR_IMG,
            RecoveredRotTrParam._value_function_handle,
            parent_parameters=parent_parameters,
        )
        self.title = "Recovered Image"
        if recovered_rot_tr_img is not None:
            self.value = recovered_rot_tr_img

    @staticmethod
    def _value_function_handle(params: Dict[Enum, Parameter]) -> np.ndarray:
        angle_select = params[RadonParams.ANGLE_SELECT].value
        return cast(
            np.ndarray, angle_select[AngleSelectParams.RECOVERED_ROT_TR_IMG].value
        )



"""
Enum of the validator.

Author: Fabian A. Preiss
"""
from enum import Enum


class ValidatorParams(Enum):
    IMG = "IMG"
    REF_IMG = "REF_IMG"
    NORM_REL_L2 = "NORM_REL_L2"
    ABSOLUTE_DIFFERENCE_IMG = "ABSOLUTE_DIFFERENCE_IMG"
    SQUARED_DIFFERENCE_IMG = "SQUARED_DIFFERENCE_IMG"



"""
Collection of useful helper functions in specific usecases

Author: Fabian A. Preiss
"""
from typing import Generator, Iterable, Sequence
import numpy as np
from PIL import Image

from imgreg.util.solver import Solver
from imgreg.util.methods import ImageMethods
from imgreg.util.io import DirectoryView
from imgreg.util.params import ParameterError


def image_file_gen(
    dview: DirectoryView, step: int = 1
) -> Generator[np.ndarray, None, None]:
    """
    Generate the images from a given DirectoryView as numpy arrays

    Parameters
    ----------
    dview : DirectoryView
        Generate images from top level folder
    step : int
        Stepsize in which files are taken
    """
    for file_path in dview.file_path_generator(step):
        yield 1.0 * np.array(Image.open(file_path))


def solver_gen(
    dview: DirectoryView, solver: Solver, step: int = 1
) -> Generator[Solver, None, None]:
    """Sweep the MOD_IMG parameters of a solver from a DirectoryView with a given stepsize.

    Parameters
    ----------
    dview : DirectoryView
        Generate images from top level folder
    solver : Solver
        Instance of a solver that
    step : int
        Stepsize in which files are taken
    """
    image_file_g = image_file_gen(dview, step)
    try:
        solver.REF_IMG.value
    except ParameterError as err:
        raise ParameterError(
            "Solver is expected to have the REF_IMG parameter set."
        ) from err
    solver.MOD_IMG.value = next(image_file_g)
    yield solver
    for arr in image_file_g:
        solver.MOD_IMG.value = arr
        yield solver


def rot_tr_gen(solvers: Iterable[Solver]) -> Generator[np.ndarray, None, None]:
    """
    Sweep over solver instances and return `tr_x` `tr_y` `tr_err` `rot` `rot_err` `NormRel_L2`

    Parameters
    ----------
    solvers : Iterable
        An iterable of solver instances
    """
    for solv in solvers:
        yield np.array(
            (
                *solv.RECOVERED_TRANSLATION.value,
                *solv.RECOVERED_ROTATION.value,
                ImageMethods.norm_rel_l2(
                    solv.RECOVERED_ROT_TR_IMG.value, solv.REF_IMG.value
                ),
            )
        )


def rot_scale_tr_gen(solvers: Iterable[Solver]) -> Generator[np.ndarray, None, None]:
    """
    Sweep over solver instances and return the rotations, scales and translations

    Parameters
    ----------
    solvers : Iterable
        An iterable of solver instances

    Returns
    -------
    np.array
        Containing `tr_x` `tr_y` `tr_err` `rot` `rot_err` `scale` `scale_err` `NormRel_L2`
    """
    for solv in solvers:
        yield np.array(
            (
                *solv.RECOVERED_TRANSLATION.value,
                *solv.RECOVERED_ROTATION.value,
                *solv.RECOVERED_SCALE.value,
                ImageMethods.norm_rel_l2(
                    solv.RECOVERED_ROT_SCALE_TR_IMG.value, solv.REF_IMG.value
                ),
            )
        )


def image_save_back_tf(
    rot_tr_arr: np.ndarray, fnames: Sequence[str], src_path: str, dest_path: str
):
    """
    Creates backtransformed images

    For an external program that compares images `*_A.*` with `*_B.*` this prepares images
    such that::

      test00001.jpg -> test00001_A.png
      test00021.jpg -> TR -> test00001_B.png, test00021_A.png

      test00221.jpg -> TR -> test00201_B.png, test00221_A.png
      test00241.jpg -> TR -> test00221_B.png

    where TR denotes the backtransformation.
    """
    for i, fname in enumerate(fnames):
        if i == 0:
            pass
        src_arr = np.array(Image.open(src_path + "/" + fname))
        dest_arr = ImageMethods.compute_rts(
            src_arr,
            angle=rot_tr_arr[i][3],
            translation=rot_tr_arr[i][:2],
            inverse=True,
        )
        dest_img = Image.fromarray(dest_arr).convert("L")
        if i != len(fnames) - 1:
            out_A = dest_path + "/" + fname.split(".")[0] + "_A.png"
            dest_img.save(out_A)
        if i != 0:
            out_B = dest_path + "/" + fnames[i - 1].split(".")[0] + "_B.png"
            dest_img.save(out_B)



"""Test images."""
from PIL import Image
import pathlib

path = pathlib.Path(__file__).parent.absolute()


def ref_img():
    return Image.open(f"{path}/ref_img.jpg")


def mod_img():
    return Image.open(f"{path}/mod_img.jpg")



"""
Enum of the radon parameters.

Author: Fabian A. Preiss
"""
from enum import Enum


class RadonParams(Enum):
    REF_IMG = "REF_IMG"
    MOD_IMG = "MOD_IMG"
    ANGULAR_PRECISION = "ANGULAR_PRECISION"
    EXPONENTIAL_FILTER_SIGNAL_NOISE = "EXPONENTIAL_FILTER_SIGNAL_NOISE"
    THETA = "THETA"
    UPSAMPLING = "UPSAMPLING"
    REF_ROTATION = "REF_ROTATION"
    MOD_ROTATION = "MOD_ROTATION"
    ROTATION_CANDIDATE = "ROTATION_CANDIDATE"
    ANGLE_SELECT = "ANGLE_SELECT"
    RECOVERED_ROTATION = "RECOVERED_ROTATION"
    RECOVERED_ROT_IMG = "RECOVERED_ROT_SCALE_IMG"
    RECOVERED_TRANSLATION = "RECOVERED_TRANSLATION"
    RECOVERED_ROT_TR_IMG = "RECOVERED_ROT_TR_IMG"



"""
Convenience functions for directed acyclic graphs

Author: Fabian A. Preiss.
"""
from typing import Optional, Callable, Dict, Set, Hashable
import itertools
from graphviz import Digraph
from imgreg.util.params import Parameter


class DAGraph:
    r"""
    Collection of functions for directed acyclic graphs.

    This class implements functions for directed acyclic graphs [#f3]_ [#f4]_.
    The graph is stored in a dictionary, where the keys define the identifiers for
    the vertices and the value contains the set of parents. This implementation
    is motivated for handling generic problems where dependencies appear (see
    dependency graphs [#f1]_).

    This module is for experimental use only, more extensive graph libraries for
    python are available under [#f2]_ [#f7]_ [#f8]_ [#f9]_.

    Parameters
    ----------
    vertex_parent_dict : dict[str, set[str]]
        A dictionary mapping each vertex to all of its parents.
        It is assumed, that the input is a directed acyclic graph (connected
        or disconnected). Cycle Detection is not performed on the input.
    invert : boolean
        invert the edges of the input graph

    Notes
    -----
    A directed graph is an ordered pair :math:`G=\left(V,E\right)`, where

    * :math:`V` is a set of vertices (also nodes or points)
    * :math:`E\subseteq\left\{ (x,y)\mid(x,y)\in V^{2}\;\textrm{ and }\;x\neq y\right\}`
      is a set of ordered pairs of vertices called edges.

    To represent a graph, DAGraph takes a single python dictionary as input, where
    :math:`\mathtt{key},\,parent\in V` and

    .. math::
        \mathtt{value}=pa\left(\mathtt{key}\right)=\left\{ \left\{ parent\right\} |
        \left(\mathtt{key},parent\right)\in E\right\}

    References
    ----------
    .. [#f3] `Wikipedia, "Graph theory" <https://en.wikipedia.org/wiki/Graph_theory>`_
    .. [#f4] `Wikipedia,
             "Directed acyclic graph" <https://en.wikipedia.org/wiki/Directed_acyclic_graph>`_
    .. [#f1] `Wikipedia, "Dependency graph" <https://en.wikipedia.org/wiki/Dependency_graph>`_
    .. [#f6] `Wikipedia, "Tree (graph theory)" <https://en.wikipedia.org/wiki/Tree_(graph_theory)>`_
    .. [#f2] https://networkx.org/
    .. [#f7] https://graph-tool.skewed.de/
    .. [#f8] https://pygsp.readthedocs.io/en/stable/
    .. [#f9] https://igraph.org/python/
    """

    def __init__(self, vertex_parent_dict: Dict[Hashable, Set[Hashable]], invert=False):
        #  NOTE: Design decision: Use a dictionary to store all ascendants
        #  for each vertex in a set
        #  Contra:
        #    - one time computational cost with constructor
        #    - additional storage requirements
        #  Pro:
        #    - simplifies implementation for certain algorithms and
        #      allows faster access of often demanded properties
        self.invert = invert
        if invert:
            tmp_dag = DAGraph(vertex_parent_dict)
            self.__vertex_parent_dict = {
                key: tmp_dag.children(key) for key in tmp_dag.vertex_parent_dict
            }
            self.__vertex_ascendants_dict = {
                key: tmp_dag.descendants(key) for key in tmp_dag.vertex_parent_dict
            }
        else:
            self.__vertex_parent_dict = vertex_parent_dict
            self.__vertex_ascendants_dict = {
                key: self.__ascendants(key) for key in vertex_parent_dict.keys()
            }

    @property
    def vertex_parent_dict(self) -> Dict[Hashable, Set[Hashable]]:
        """
        Get a dictionary mapping each vertex to all of its parents.

        Returns
        -------
        dict
            vertex : Set of vertices
        """
        return self.__vertex_parent_dict

    @property
    def vertex_ascendants_dict(self) -> Dict[Hashable, Set[Hashable]]:
        """
        Get a dictionary mapping each vertex to all of its ascendants [#f6]_.

        Returns
        -------
        dict
            vertex : Set of vertices
        """
        return self.__vertex_ascendants_dict

    def __r_ascendants(
        self,
        vertex: Hashable,
        tested: Set[Hashable] = None,
        collected: Set[Hashable] = None,
    ):
        """
        Recursively ascend the graph and collect all ascendants of *vertex*.

        The final set contains all vertices from which *vertex* is
        reachable[#f5].
        Recursion overwrites *tested* and *collected*.

        Parameters
        ----------
        vertex : Hashable
            a vertex of the graph
        tested : Set
            Set of vertices tested for ascendants
        collected : Set
            Set of vertices that have been tagged as ascendants so far
        """
        vertex_parent_dict = self.__vertex_parent_dict
        tested = set() if tested is None else tested
        tested.update({vertex})
        collected = set() if collected is None else collected
        collected.update(vertex_parent_dict[vertex])
        for dep in vertex_parent_dict[vertex]:
            if dep not in tested and dep is not None:
                self.__r_ascendants(dep, tested=tested, collected=collected)

    def __ascendants(self, vertex: Hashable) -> Set[Hashable]:
        """
        Given *vertex* returns the set of all vertices from which *vertex* is reachable[#f5].

        .. [#f5] https://en.wikipedia.org/wiki/Reachability

        Parameters
        ----------
        vertex : Hashable
            a vertex of the graph

        Returns
        -------
        Set
            Set of vertices that are ascendants from vertex
        """
        res: Set[Hashable] = set()
        self.__r_ascendants(vertex, collected=res)
        return res

    def parents(self, vertex: Hashable, order=1) -> Set[Hashable]:
        """
        Given vertex, return a set containing all its parents.

        Returns
        -------
        Set
            Set of *parent*
        """
        current_parents = self.__vertex_parent_dict[vertex]
        if order > 1:
            return set(
                itertools.chain(
                    *(self.parents(parent, order - 1) for parent in current_parents)
                )
            )
        return current_parents

    def ascendants(self, vertex: Hashable) -> Set[Hashable]:
        """
        Given vertex, return a set containing all its ascendants [#f6]_.

        Returns
        -------
        Set
            Set of *ascendants*
        """
        return self.__vertex_ascendants_dict[vertex]

    def children(self, vertex: Hashable, order: int = 1) -> Set[Hashable]:
        """
        Given vertex, return a set containing all its children.

        Returns
        -------
        Set
            Set of *children*
        """
        current_children = {
            key
            for key in self.__vertex_parent_dict
            if vertex in self.__vertex_parent_dict[key]
        }
        if order > 1:
            return set(
                itertools.chain(
                    *(self.children(child, order - 1) for child in current_children)
                )
            )
        return current_children

    def descendants(self, vertex: Hashable) -> Set[Hashable]:
        """
        Given vertex, return a set containing all descendants [#f6]_.

        Returns
        -------
        Set
            Set of *descendants*
        """
        return {
            key
            for key in self.__vertex_ascendants_dict
            if vertex in self.__vertex_ascendants_dict[key]
        }
